"""
Module to handle universal/general constants used across files.
"""

################################################################################
# Constants #
################################################################################

# GENERAL CONSTANTS:
VERY_SMALL_NUMBER = 1e-31
INF = 1e20


_PAD_TOKEN = '#pad#'
_UNK_TOKEN = '<unk>'
_SOS_TOKEN = '<s>'
_EOS_TOKEN = '</s>'


# LOG FILES ##
_CONFIG_FILE = 'config.json'
_SAVED_WEIGHTS_FILE = 'params.saved'
_SAVED_METRICS_FILE = 'metric.log'
# _PREDICTION_FILE = "test_pred.txt"
# _REFERENCE_FILE = "test_ref.txt"
